// Copyright 2018 Orus.io. All rights reserved.

package sql

import (
	"database/sql"
	"fmt"

	"github.com/lib/pq"
	"github.com/m4rw3r/uuid"
	"github.com/orus-io/yago"
	"github.com/rs/zerolog"

	"xbus.io/go-xbus/v3/api"
	"xbus.io/xbus/storage"
	"xbus.io/xbus/storage/sql/model"
)

type rowScanner interface {
	Scan(...interface{}) error
}

const (
	insertFieldClause  = "process_id, actor_id, node_id, created_at, detached, status, envelope_ids"
	insertValuesClause = "$1, $2, $3, $4, $5, $6, $7"

	allFieldClause = insertFieldClause + ", firstattempt_at, attempt_count"

	insertQuery            = "INSERT INTO job (" + insertFieldClause + ") VALUES (" + insertValuesClause + ")"
	selectQuery            = "SELECT " + allFieldClause + " FROM job"
	primaryKeyFilterClause = "process_id = $1 AND actor_id = $2 and node_id = $3"
	updateStateQuery       = `UPDATE job
	SET detached = $4, status = $5, envelope_ids = $6, firstattempt_at = $7, attempt_count = $8
	WHERE ` + primaryKeyFilterClause
)

func scanJob(scanner rowScanner, job *storage.Job) error {
	var (
		statusString   string
		envelopeIDs    []uuid.UUID
		firstAttemptAt sql.NullTime
	)
	if err := scanner.Scan(
		(*uuid.UUID)(&job.ProcessID),
		(*uuid.UUID)(&job.Node.ActorID),
		&job.Node.NodeID,
		&job.CreatedAt,
		&job.Detached,
		&statusString,
		pq.Array(&envelopeIDs),
		&firstAttemptAt,
		&job.AttemptCount,
	); err != nil {
		return err
	}
	if err := job.Status.UnmarshalText([]byte(statusString)); err != nil {
		return err
	}
	if len(envelopeIDs) != 0 {
		job.EnvelopeIDs = make([]api.UUID, len(envelopeIDs))
		for i, id := range envelopeIDs {
			job.EnvelopeIDs[i] = api.UUID(id)
		}
	}
	if firstAttemptAt.Valid {
		job.FirstAttempt = firstAttemptAt.Time
	}
	return nil
}

func jobValues(job storage.Job, insert bool) []interface{} {
	statusBytes, err := job.Status.MarshalText()
	if err != nil {
		panic(err)
	}
	envelopeIDs := make([]uuid.UUID, len(job.EnvelopeIDs))
	for i, id := range job.EnvelopeIDs {
		envelopeIDs[i] = uuid.UUID(id)
	}
	var firstAttemptAt sql.NullTime
	if !job.FirstAttempt.IsZero() {
		firstAttemptAt.Time = job.FirstAttempt.UTC()
		firstAttemptAt.Valid = true
	}
	var values = []interface{}{
		uuid.UUID(job.ProcessID),
		uuid.UUID(job.Node.ActorID),
		job.Node.NodeID,
	}
	if insert {
		values = append(values, job.CreatedAt.UTC())
	}
	values = append(values,
		job.Detached,
		string(statusBytes),
		pq.Array(envelopeIDs),
	)
	if !insert {
		values = append(values,
			firstAttemptAt,
			job.AttemptCount,
		)
	}
	return values
}

// NewJobStorage initialize a new JobStorage
func NewJobStorage(db *yago.DB, logger zerolog.Logger) *JobStorage {
	return &JobStorage{
		Base: Base{
			model: nil,
			db:    db,
			logger: logger.With().
				Str("storage", "job").
				Str("implementation", "sql").
				Logger(),
		},
	}
}

// JobStorage implements storage.JobStorage
type JobStorage struct {
	Base
}

// NewJob ...
func (s *JobStorage) NewJob(
	id storage.JobID, state storage.JobState,
) {
	tx := s.MustBegin()
	defer tx.RollbackIfOpened()

	_, err := tx.GetEngine().Exec(
		model.RawSQL(
			insertQuery,
			jobValues(storage.Job{JobID: id, JobState: state}, true)...,
		))
	if err != nil {
		panic(err)
	}

	tx.MustCommit()
}

// GetState returns the job state
func (s *JobStorage) GetState(id storage.JobID) storage.JobState {
	tx := s.MustBegin()
	defer tx.BlindRollback()

	var job storage.Job

	row := tx.GetEngine().QueryRow(
		model.RawSQL(
			selectQuery+" WHERE "+primaryKeyFilterClause,
			uuid.UUID(id.ProcessID), uuid.UUID(id.Node.ActorID), id.Node.NodeID,
		))
	must(scanJob(row, &job))

	return job.JobState
}

// SetState ...
func (s *JobStorage) SetState(id storage.JobID, state storage.JobState) {
	tx := s.MustBegin()
	defer tx.RollbackIfOpened()

	_, err := tx.GetEngine().Exec(model.RawSQL(
		updateStateQuery, jobValues(storage.Job{JobID: id, JobState: state}, false)...,
	))
	must(err)

	tx.MustCommit()
}

type jobList struct {
	tx   *TxHelper
	rows *sql.Rows
}

func (l jobList) Next() bool {
	return l.rows.Next()
}

func (l jobList) Get() (storage.Job, error) {
	var job storage.Job
	err := scanJob(l.rows, &job)
	return job, err
}

func (l jobList) Close() {
	l.rows.Close()
	l.tx.BlindRollback()
}

func (s *JobStorage) streamList(orderby string, where string, bindings ...interface{}) (
	storage.JobList, error,
) {
	tx := s.MustBegin()

	var query = selectQuery
	if where != "" {
		query += " WHERE " + where
	}
	if orderby != "" {
		query += " ORDER BY " + orderby
	}

	rows, err := tx.GetEngine().Query(
		model.RawSQL(query, bindings...))
	if err != nil {
		return nil, err
	}

	return jobList{tx, rows}, nil
}

func (s *JobStorage) list(orderby string, where string, bindings ...interface{}) []storage.Job {
	list, err := s.streamList(orderby, where, bindings...)
	must(err)

	defer list.Close()

	var jobs []storage.Job
	for list.Next() {
		job, err := list.Get()
		must(err)
		jobs = append(jobs, job)
	}

	return jobs
}

// ListJobByStatus ...
func (s *JobStorage) ListJobByStatus(
	status api.ActorProcessingState_Status,
) []storage.Job {
	b, err := status.MarshalText()
	must(err)
	return s.list("id", "status = '$1'", string(b))
}

// ListPendingJobByActor returns the active jobs of a given actor
func (s *JobStorage) ListPendingJobByActor(
	actorID api.UUID, after *storage.JobID, limit int,
) []storage.Job {
	where := "actor_id = $1 and status = 'PENDING'"
	bindings := []interface{}{uuid.UUID(actorID)}

	if after != nil {
		where += " AND id > (SELECT id FROM job WHERE process_id = $2 AND actor_id= $3 AND node_id = $4)"
		bindings = append(bindings,
			uuid.UUID(after.ProcessID),
			uuid.UUID(after.Node.ActorID),
			after.Node.NodeID,
		)
	}
	list := s.list(fmt.Sprintf("id LIMIT %d", limit), where, bindings...)
	// TODO add the limit in the SQL query
	if len(list) > limit {
		list = list[:limit]
	}
	return list
}

// ListJobByProcess returns the active jobs of a given process
func (s *JobStorage) ListJobByProcess(processID api.UUID) []storage.Job {
	return s.list(
		"id",
		"process_id = $1 AND status IN ($2, $3)",
		uuid.UUID(processID),
		"PENDING",
		"RUNNING")
}

// ListRunningJobByEnvelopeID returns the active jobs that references a given envelope
func (s *JobStorage) ListRunningJobByEnvelopeID(envelopeID api.UUID) []storage.Job {
	return s.list(
		"id",
		"$1 = ANY(envelope_ids) AND status = $2",
		uuid.UUID(envelopeID),
		"RUNNING")
}

// ListDetachedJobByActor ...
func (s *JobStorage) ListDetachedJobByActor(actorID api.UUID) (storage.JobList, error) {
	return s.streamList(
		"id",
		"actor_id=$1 AND detached = 't' AND status = $2",
		uuid.UUID(actorID),
		"RUNNING",
	)
}

// Purge removes jobs
func (s *JobStorage) Purge(processIDs []api.UUID) error {
	tx := s.MustBegin()
	defer tx.RollbackIfOpened()

	var dbids = make([]uuid.UUID, len(processIDs))
	for i, id := range processIDs {
		dbids[i] = uuid.UUID(id)
	}

	_, err := tx.GetEngine().Exec(model.RawSQL(
		"DELETE FROM job WHERE process_id = ANY($1)",
		pq.Array(dbids)))
	if err == nil {
		tx.MustCommit()
	}
	return err
}
